function [X,D,res,cycle,nmatvec]=even_ira(M,N,shift,nwant,opts)
%even_ira - implicitly restarted Arnoldi method for even eigenvalue problems
%
% use:
%  [...] = even_ira(M,N,shift)
%  [...] = even_ira(M,N,shift,nwant)
%  [...] = even_ira(M,N,shift,nwant,opts)
%
%  e                   = even_ira(...)
%  [X,D]               = even_ira(...)
%  [X,D,cycle]         = even_ira(...)
%  [X,D,cycle,nmatvec] = even_ira(...)
%
% inputs:
%  M - real symmetric matrix, size n-by-n
%  N - real skew symmetric matrix, size n-by-n
%      eigenvalues of M*x=lambda*N*x are wanted
%  shift - eigenvalues near shift are wanted
%      shift must be a real or purely imaginary scalar
%  nwant - number of wanted eigenvalues (default: 6)
%  opts - structure containing additional options, default values are used
%      for unprovided options, recognized fields are:
%  opts.tol- tolerance, a pair (lambda,x) is considered converged if
%      norm(M*x-lambda*N*x) <= opts.tol, (default: 10^(-6))
%  opts.dimmax - maximal search space dimension, must be >= opts.dimrestart+2
%      larger values are better, but need more memory
%      (default: max(2*opts.dimrestart,20))
%  opts.dimrestart - integer, number of vectors to restart with,
%      must be larger or equal ceil(nwant/2) (default: ceil(nwant/2)+1)
%  opts.nrestart - integer, nonnegative, max number of restarts (default: 10)
%  opts.v1 - real vector, size n, start vector for Arnoldi iteration,
%      see also opts.powerfirst (default: randn(n,1))
%  opts.powerfirst - integer, preprocessing for startvector using power iteration
%      0 -> use v as start vector
%      1 -> use inv(M+sigma*N)*N*v
%      2 -> use inv(M-sigma*N)*N*v
%      3 -> use inv(M+sigma*N)*N*inv(M-sigma*N)*N*v
%      for imaginary shift, only values 0 and 3 are permitted
%      nonzero value is recommended for singular or ill-conditioned N
%      (default: 3)
%  opts.nitref - integer, defining, how to solve linear systems
%      0 -> use lu-factorization without iterative refinement
%      k -> use lu-factorization with k steps of iterative refinement (k>0)
%     -1 -> use backslash for every linear system (slow)
%      (default: 1)
%  opts.buildNV - boolean, if true, an orthonormal basis for N*searchspace is built;
%      if false, non-orthogonal orthogonalization is used instead;
%      using opts.buildNV=false halves the memory requirements (compared to opts.buildNV=true),
%      but is not recommended if N is ill conditioned or singular
%      (default: buildNV=true)
%  opts.OPform - integer, defining how the operator 
%      OP x:=inv(M+shift*N)*N*inv(M-shift*N)*N*x is applied,
%      0 -> (inv(M+shift*N)*(N*(inv(M-shift*N)*(N*x)))),
%      1 -> (inv(M-shift*N)*N*x-inv(M+shift*N)*N*x)/(2*shift),
%      2 -> imag(inv(M-shift*N)*N*x)/imag(shift),
%      value 1 requires nonzero shift,
%      value 2 requires nonzero, purely imaginary shift,
%      all three formulations are equivalent in exact arithmetic,
%      values 1 and 2 may suffer from cancellation
%      value 2 is more efficient (only one inverse involved)
%      (default: 2 (for nonzero, purely imaginary shift), or 0 (oterwise) )
%  opts.disp - integer, defining printout of information
%      0 -> silent mode, no information is printed
%      1 -> basic information is printed
%      2 -> more information is printed
%      (default: 1)
%
% outputs:
%  e - vector, length nconv, contains found eigenvalues
%      where nconv is the number of converged eigenvalues,
%      the eigenvalues of even problems come in (real or purely imaginary)
%      pairs, lambda,-lambda, or in (non-real,non-purely-imaginary)
%      quadruples, lambda,-lambda,conj(lambda), -conj(lambda);
%      care is taken to return pairs and quadruples as a whole,
%      nconv should be close to nwant, but can be larger (to capture
%      whole pairs/quadruples) or smaller (maximum number of iteraton have
%      been taken, but not enough eigenpairs converged),
%  X - matrix size n-by-nconv, contains eigenvectors,
%  D - matrix, size nconv-by-nconv, contains found eigenvalues on diagonal
%      it holds: M*X is close to N*X*D
%  res - vector, length nconv, norms of the residuals
%      res(i)=norm(M*X(:,i)-D(i,i)*N*X(:,i))
%  cycle - integer, number of used arnoldi-cycles
%  nmatvec - integer, number of used operator-vector multiplications, with
%          operator = inv(M-shift*N)*N or inv(M+shift*N)*N
%      if nitref=-1 -> the number of backlsah operations (regardless of the
%          number of right hand sides)
%      if nitref>=0 -> the number of vectors the lu factorization is
%          applied to

% method/reference:
%  Even implicitly restarted Arnoldi method, as described in
%
%  V. Mehrmann, C. Schroeder, and V. Simoncini,
%  An Implicitly-restarted Krylov Method for Real Symmetric/Skew-Symmetric Eigenproblems,
%  Linear Algebra and its Applications, vol. 436, pp. 4070--4087, 2012

% license:
% The MIT License (MIT)
% 
% Copyright (c) 2013 Volker Mehrmann, Valeria Simoncini, Christian Schroeder
% 
% Permission is hereby granted, free of charge, to any person obtaining a copy of
% this software and associated documentation files (the "Software"), to deal in
% the Software without restriction, including without limitation the rights to
% use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
% the Software, and to permit persons to whom the Software is furnished to do so,
% subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included in all
% copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
% IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
% FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
% COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
% IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
% CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

%check parameters
if nargin<3,error('Not enough input arguments');end

%check M,N,shift
n=size(M,1);
if any([size(M),size(N)]~=n)
    error('M and N must be square of the same size');
end
if any(any(imag(M))) || ~isequal(M,M.')
    error('M must be real symmetric');
end
if any(any(imag(N))) || ~isequal(N,-N.')
    error('N must be real and skew symmetric');
end
if numel(shift)~=1 || (real(shift)~=0 && imag(shift)~=0)
    error('shift must be a real or purely imaginary scalar');
end

%check nwant, opts
if nargin<4 || isempty(nwant),
    nwant=6;
else
    if nwant<0,
        error('nwant must be positive');
    end
end

if nargin<5 || isempty(opts),
    opts=struct;%empty structure
end

%process options
if isfield(opts,'tol')
    if opts.tol<0
        error('opts.nrestart must be nonnegative');
    end
else
    opts.tol=1e-6;
end
if isfield(opts,'dimrestart')
    if opts.dimrestart<ceil(nwant/2)
        error('opts.dimrestart must be at least ceil(nwant/2)');
    end
else
    opts.dimrestart=ceil(nwant/2)+1;
end
if isfield(opts,'dimmax')
    if opts.dimmax<opts.dimrestart+2
        error('opts.dimmax must be at least opts.dimrestart+2');
    end
else
    opts.dimmax=min(max(2*opts.dimrestart,20),n);
end
if isfield(opts,'nrestart')
    if opts.nrestart<0
        error('opts.nrestart must be nonnegative');
    end
else
    opts.nrestart=10;
end
if isfield(opts,'buildNV')
    %nothing
else
    opts.buildNV=true;
end
if isfield(opts,'nitref')
    %nothing
else
    opts.nitref=1;
end
if isfield(opts,'powerfirst')
    if all(opts.powerfirst~= 0:3)
        error('opts.powerfirst must be one of {0,1,2,3}');
    end
    if imag(shift)~=0 && opts.powerfirst~=0 && opts.powerfirst~=3
        error('opts.powerfirst must be one of {0,3} for imaginary shift');
    end
else
    opts.powerfirst=3;
end
if isfield(opts,'v1')
    v=opts.v1;
    if any(size(v)~=[n,1])
        error('opts.v1 must be an n-dim column vector');
    end
    if any(imag(v))
        error('opts.v1 must be real');
    end
else
    v=randn(n,1);
end
if isfield(opts,'OPform')
    if all(opts.OPform~= 0:2)
        error('opts.OPform must be one of {0,1,2}');
    end
    if shift==0 && opts.OPform==1
        error('opts.OPform = 1 is not permitted for zero shift');
    end
    if imag(shift)==0 && opts.OPform==2
        error('opts.OPform = 2 requires nonzero, purely imaginary shift');
    end
else
    if imag(shift)~=0
        opts.OPform=2;
    else
        opts.OPform=0;
    end
end
if isfield(opts,'disp')
    if all(opts.disp~= 0:2)
        error('opts.disp must be one of {0,1,2}');
    end
else
    opts.disp=1;
end

%disp setting
if opts.disp>0
    fprintf('\nThis is Even-Ira, the implicitly restarted Arnoldi method\n');
    fprintf('for even eigenvalue problems, version 20130815\n\n');
    fprintf('n: %i, nwant: %i, v1: ',n,nwant);
    if isfield(opts,'v1'), fprintf('provided\n');else fprintf('random\n');end
    fprintf('shift: ');disp(shift);
    fprintf('options:\n')
    disp(opts)
end

%allocate matrices
V=zeros(n,opts.dimmax);%search space
H=zeros(opts.dimmax,opts.dimmax-1);
nmatvec=0;
if opts.buildNV,
    NV=zeros(n,opts.dimmax);%orthogonal basis for N*V
else
    Z=zeros(opts.dimmax);%stores V'*N'*N*V
end

%setup shift-invert-operator
if opts.nitref >=0,
    if opts.disp>0,fprintf('perform lu factorization... ');end
    try
        [L,U,pp,qq,dd]=lu(M-shift*N,'vector');
    catch%call to lu did not work (possibly because of old matlab version)
        disp('Try old matlab lu syntax');
        [L,U,pp,qq]=lu(M-shift*N);
        %transform permutation matrices to permutation vectors
        pp=(1:n)*pp';
        qq=(1:n)*qq;
        dd=speye(n);%identity scaling 
    end
    if opts.disp>0,fprintf('done\n');end
else
    L=0;U=0;pp=0;qq=0;dd=0;%variables must exist, but are not used
end

%preprocessing starting vector
%one step of power method to project out directions in eigenvectors
%cooresponding to infinite eigenvalues
v=real(v);
switch opts.powerfirst
    case 0
        %do nothing
    case 1
        [v,nmatvec]=applyinvMplusshiftNN(v,M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec);
    case 2
        [v,nmatvec]=applyinvMminusshiftNN(v,M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec);
    case 3
        [v,nmatvec]=applyinvMplusminusshiftNN(v,M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec,opts.OPform);
end
v=real(v);

V(:,1)=v/norm(v);
if opts.buildNV,
    NV(:,1)=N*V(:,1);
    NV(:,1)=NV(:,1)/norm(NV(:,1));
else
    w=N*V(:,1);
    Z(1,1)=w'*w;
end
m=1;%V is n-by-m, H is m-by-(m-1)
nconv=0;%number of converged eval-pairs
cycle=0;%number of current Arnoldi cycle
while nconv<nwant && cycle<=opts.nrestart,
    cycle=cycle+1;
    while m<opts.dimmax,%Expand search space via Arnoldi
        %apply operator: w= inv(M+shift*N)'*N*inv(M-shift*N)*N*V(:,m);
        [w,nmatvec]=applyinvMplusminusshiftNN(V(:,m),M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec,opts.OPform);
        %orthogonalization vs V, NV
        for ii=1:2,
            %NV-orthogonalization, in exact arithmetic, this
            %automatically holds, in finite arithmetic we make sure.
            if opts.buildNV,
                for i=1:m,
                    gamma=NV(:,i)'*w;
                    w=w-gamma*NV(:,i);
                end
            else
                gamma=-(V(:,1:m)'*(N*w));
                gamma=Z(1:m,1:m)\gamma;
                w=w-N*(V(:,1:m)*gamma);
            end
            for i=1:m,%orthogonalization vs V
                gamma=V(:,i)'*w;
                w=w-gamma*V(:,i);
                H(i,m)=H(i,m)+gamma;
            end
        end
        H(m+1,m)=norm(w);
        w=w/H(m+1,m);
        V(:,m+1)=w;
        w=N*w;
        if opts.buildNV,
            for ii=1:2,
                for i=1:m,
                    gamma=NV(:,i)'*w;
                    w=w-gamma*NV(:,i);
                end
            end
            NV(:,m+1)=w/norm(w);
        else
            Z(m+1,m+1)=w'*w;
            Z(1:m,m+1)=-V(:,1:m)'*(N*w);
            Z(m+1,1:m)=Z(1:m,m+1)';
        end
        m=m+1;
    end%end arnoldi expansion

    %implicit restart (stewart style)
    [Q,H(1:m-1,1:m-1)]=schur(H(1:m-1,1:m-1));
    e=ordeig(H(1:m-1,1:m-1));
    [sae,p]=sort(abs(e),1,'descend');
  
    %determining number of eigenvalues to keep
    %should be opts.dimrestart, but is increased, if there are more evals
    %of same modulus (eg. because of complex pair)
    nkeep=opts.dimrestart;
    while nkeep<m-1 && sae(nkeep+1)==sae(nkeep),
        nkeep=nkeep+1;
    end
    ip=zeros(m-1,1); ip(p)=1:m-1;%inverse permutation
    cluster= (ip<=nkeep);%kept evals are unordered
    %cluster= max(nkeep+1-ip,0);%kept evals are ordered
    [Q,H(1:m-1,1:m-1)]=ordschur(Q,H(1:m-1,1:m-1),cluster);
    %truncation
    H(nkeep+1,1:nkeep)=H(m,m-1)*Q(m-1,1:nkeep);
    H(:,nkeep+1:opts.dimmax-1)=0;%the now-unused part of H is set to zero, because Gram-Schmidt works additive
    H(nkeep+2:opts.dimmax,1:nkeep)=0;
    V(:,1:nkeep)=V(:,1:m-1)*Q(1:m-1,1:nkeep);
    V(:,nkeep+1)=V(:,m);%the now-unused part of V, does not have to be set to zero, because its overwritten later
    if opts.buildNV,
        [NV(:,1:nkeep+1),junk]=qr(N*V(:,1:nkeep+1),0);
    else
        Z(1:nkeep+1,1:nkeep+1)=blkdiag(Q(:,1:nkeep),1)'*...
            Z(1:m,1:m)*blkdiag(Q(:,1:nkeep),1);
        Z(1:nkeep+1,1:nkeep+1)=...
            (Z(1:nkeep+1,1:nkeep+1)+Z(1:nkeep+1,1:nkeep+1)')/2;
    end
    m=nkeep+1;
    
    %getting residuals of original  problem
    [W1,nmatvec]=applyinvMminusshiftNN(V(:,1:m-1),M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec);
    if real(shift)==0,
        W2=conj(W1);
    else
        [W2,nmatvec]=applyinvMplusshiftNN(V(:,1:m-1),M,N,shift,L,U,pp,qq,dd,opts.nitref,nmatvec);
    end
    
    [S,mu]=eig(H(1:m-1,1:m-1));
    mu=full(diag(mu));mu=mu(:); %mu is now a column vector
    [junk,p]=sort(abs(mu),1,'descend');
    mu=mu(p);S=S(:,p);
    idx=[false;mu(2:end)==conj(mu(1:end-1))];% the indices of the mu's that are conjugates of the previous mu
    idxn=~idx;
    idx=find(idx);idxn=find(idxn);
    la=zeros(m-1,1);
    la(idxn)=sqrt(1./mu(idxn)+shift^2);
    la(idx)=conj(la(idx-1));
    
    idx= (abs(shift-la)>abs(shift+la)); la(idx)=-la(idx);%take square root closer to shift

    %form (approx) eigenvalues and -vectors
    X=zeros(n,2*(m-1));
    X(:,2:2:2*(m-1)) =V(:,1:m-1)*S-W2*S*diag(sparse(la+shift));
    X(:,1:2:2*(m-1)) =V(:,1:m-1)*S+W1*S*diag(sparse(la+shift));
    X=X/diag(sparse(sqrt(sum(abs(X).^2))));
    la=la(kron(1:m-1,[1,1]),1); la(2:2:2*m-2)=-la(2:2:2*m-2);% [la1, -la1, la2, -la2, ...]
    D=diag(sparse(la));

    %compute residual
    R=M*X-(N*X)*D;
    res=sqrt(sum(abs(R).^2));

    %counts, how many of the ritzvalues closest to the shift have converged
    %make sure that pairs/quadruples of eigenvalues converge as a whole
    nconv=0;
    while nconv+1<=length(res),
        maxresblock=res(nconv+1);
        blocksize=1;
        while nconv+blocksize+1<=length(res) ...
                && abs(real(la(nconv+blocksize+1)))==abs(real(la(nconv+1)))...
                && abs(imag(la(nconv+blocksize+1)))==abs(imag(la(nconv+1)))
            blocksize=blocksize+1;
            maxresblock=max(maxresblock,res(nconv+blocksize));
        end
        if maxresblock<=opts.tol
            nconv=nconv+blocksize;
        else
            break;
        end
    end
    
    % display some infos
    if opts.disp>=2
        fprintf('\ncycle: %i, converged eigenpairs: %i\n',cycle,nconv);
        fprintf(' approx.eigs (real,imag),  log10 of residuals, converged?\n');
        fprintf(' %12.5e + %12.5ei, %5.1f, %i\n',[real(la(:)),imag(la(:)),log10(res(:)),(1:2*(m-1))'<=nconv].');
    end
end%cycles

%truncate outputs to converged eigenpairs
X=X(:,1:nconv);
D=D(1:nconv,1:nconv);
la=la(1:nconv);
res=res(1:nconv);

% special case when only eigenvalues are wanted
if nargout<=1, X=la(:);end

% display some infos
if opts.disp>0
    fprintf('\nfinished: converged eigenpairs: %i, used restarts: %i, solved linear systems: %i\n',nconv,cycle-1,nmatvec);
    fprintf(' eigs (real,imag)  and log10 of residuals\n');
    fprintf(' %12.5e + %12.5ei, %5.1f\n',[real(la(:)),imag(la(:)),log10(res(:))].');
end
end %even_ira




% auxiliary routines
function [x,nmatvec]=applyinvMminusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec)
x=N*x;
if nitref<0,
    x=(M-shift*N)\x;
    nmatvec=nmatvec+1;
else
    b=x;
    x=dd\x;
    x(qq,:)=U\(L\x(pp,:));
    nmatvec=nmatvec+1;
    for i=1:nitref,
        c=b-(M*x-shift*N*x);
        c=dd\c;
        c(qq,:)=U\(L\c(pp,:));
        x=x+c;
        nmatvec=nmatvec+1;
    end
end
end %applyinvMminusshiftNN

function [x,nmatvec]=applyinvMplusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec)
x=N*x;
if nitref<0,
    x=(M+shift*N)\x;
    nmatvec=nmatvec+1;
else
    b=x;
    x(pp,:)=L.'\(U.'\x(qq,:));
    x=dd\x;
    nmatvec=nmatvec+size(x,2);
    for i=1:nitref,
        c=b-(M*x+shift*N*x);
        c(pp,:)=L.'\(U.'\c(qq,:));
        c=dd\c;
        x=x+c;
        nmatvec=nmatvec+size(x,2);
    end
end
end %applyinvMplusshiftNN

function [x,nmatvec]=applyinvMplusminusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec,formulation)
if formulation==0
    [x,nmatvec]=applyinvMminusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec);
    [x,nmatvec]=applyinvMplusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec);
    x=real(x);
elseif formulation==1%for nonzero shift only
    [y,nmatvec]=applyinvMminusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec);
    [z,nmatvec]=applyinvMplusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec);
    x=(y-z)/(2*shift);
    x=real(x);
elseif formulation==2%for nonzero imagiary shift only
    [x,nmatvec]=applyinvMminusshiftNN(x,M,N,shift,L,U,pp,qq,dd,nitref,nmatvec);
    x=imag(x)/imag(shift);
end
end






